/*
Copyright [2020] [Alex Sorkin]

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

class colorCode {
    constructor(dom = document,newColors) {
        let colors = {inner:'white',tag:'#FD2671',prop:'#A2D529',text:'#F2E679',comment:'#6F7463',fn:'#A2D529',let:'#69E6E3',false:'#B282F2',this:'#FF971F'}
        if(newColors !== undefined) {
            for(let color in newColors) {
                if(colors[color] !== undefined) colors[color] = newColors[color]
            }
        }
        this.colors = colors
        this.actions = ['onafterprint','onbeforeprint','onbeforeunload','onerror','onhashchange','onload','onmessage','onoffline','ononline','onpagehide','onpageshow','onpopstate','onresize','onstorage','onunload','onblur','onchange','oncontextmenu','onfocus','oninput','oninvalid','onreset','onsearch','onselect','onsubmit','onclick','ondblclick','onmousedown','onmousemove','onmouseout','onmouseover','onmouseup','onmousewheel','onwheel','ondrag','ondragend','ondragenter','ondragleave','ondragover','ondragstart','ondrop','onscroll','oncopy','oncut','onpaste','onabort','oncanplay','oncanplaythrough','oncuechange','ondurationchange','onemptied','onended','onerror','onloadeddata','onloadedmetadata','onloadstart','onpause','onplay','onplaying','onprogress','onratechange','onseeked','onseeking','onstalled','onsuspend','ontimeupdate','onvolumechange','onwaiting','ontoggle',];
        this.styleSheet = `
        [inner-colorcode] { color:white;}
        [tag-colorcode] { color:${colors.tag};}
        [prop-colorcode] { color:${colors.prop};}
        [text-colorcode] { color:${colors.text};}
        [comment-colorcode] { color:${colors.comment};}
        [fn-colorcode] { color:${colors.fn};}
        [let-colorcode] { color:${colors.let};}
        [false-colorcode] { color:${colors.false};}
        [this-colorcode] { color:${colors.this};}
        `
        let codes = this.getDom(dom)
        for(let i = codes.length-1; i>=0; i--) {
            let inner = codes[i].innerHTML
            this.setProps(codes[i])
            let outer = this.router(inner.trim())
            outer = this.outerCode(outer,codes[i],i)
            if(i == 0) codes[i].insertAdjacentHTML('beforebegin',`<style>${this.styleSheet}</style>`)
            codes[i].insertAdjacentHTML('afterend',outer)
            if(this.resultProp !== undefined) this.createResult(inner,`__result${i}__`) 
            codes[i].remove()
        }
    }

    setProps(dom) {
        this.dom = dom
        let propsArray = ['html','js','css','result']
        this.propOrder = []
        let props = dom.attributes;
        for(let i in props) {
            if(typeof props[i] == 'object') {
                if(propsArray.includes(props[i].nodeName)) this.propOrder.push(props[i].nodeName)
                this[props[i].nodeName+'Prop'] = props[i].nodeValue;
            }
        }
    }

    getDom(dom) {
        if(typeof dom !== 'object') dom = document.getElementById(dom)
        if(dom.tagName == 'TEMPLATE') return [dom]
        return dom.querySelectorAll('[colorcode]')
    }

    createResult(inner,mainDomId) {
    // Create iframe
        let dom = document.getElementById(mainDomId)
        dom.innerHTML = inner
    // Set width and height
        if(!this.resultProp.includes('full')) {
            let childNodes = dom.childNodes
            let length
            let maxLength = 0
            for(let i=0; i<=childNodes.length; i++) {
                if(childNodes[i] !== undefined) {
                    if(childNodes[i].innerHTML !== undefined) length = childNodes[i].innerHTML.length
                    else length = childNodes[i].length
                    if(length>maxLength) maxLength = length
                }
            }
            if(maxLength >300) dom.style.width = '600px'
            else dom.style.width = maxLength*8+5+'px'
        // Set height
            if(dom.clientHeight > 300) dom.style.height = '300px'
        }
    // Eval scripts
        let scripts = dom.getElementsByTagName('script')
        if(scripts.length !== 0) {
            for(let i = scripts.length-1; i>=0; i--) {
                let script = document.createElement('SCRIPT')
                script.innerHTML = scripts[i].innerHTML 
                let id = Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 9)+'script';
                scripts[i].insertAdjacentHTML('afterend',`<div id="${id}"></div>`)
                document.getElementById(id).appendChild(script)
                scripts[i].remove()
            }
        }
    }

    outerCode(outer,dom,i) {
// Define styles and functions
        this.styleSheet += `\n
        [mainBoxStyle] {
            display: inline-flex;
            flex-wrap:wrap;
            flex-direction:row;
            background-color: #303030;
            border-radius: 5px;
            border-bottom:solid green 5px; 
            border-radius:5px;
            Justify-content:flex-start;
            padding:.5rem;
            width:100%;
        }
        [tabStyles]{ background-color: #303030; cursor: pointer; font-size:1.5rem; border-bottom:solid gray 1px;}
        [tabStyle]{ color: #A2D529; background-color: #303030; cursor: pointer; font-size:1.5rem; border-bottom:solid gray 1px;}
        [tabActiveStyle]{ color: #FD2671; background-color: #303030; cursor: pointer; font-size:1.5rem; border-bottom:solid gray 1px;}
        [contentStyle]{ max-width:100%; border: #303030 solid 1px;color: white;padding: .5rem;padding-right: .2rem; display:flex; flex-direction:column;}
        [resultStyle]{ background-color:white;color:black;font-size:1rem; flex-grow:1; resize: both; overflow:auto;}

        `
        let fuShowHide = `
        let content = this.parentNode.children[1]
        if(content.style.display == 'none') {
            content.style.display = 'block'
            this.style.color = '#FD2671'
        }
        else {
            content.style.display = 'none'
            this.style.color = '#A2D529'
        }`
        let hideStyle = ''
        let active
        let size
        let tabs1 = `<div mainBoxStyle>`

        if(this.propOrder.length > 0) {
            let props = this.propOrder
            this.html = outer
            props.forEach((prop,index) => {
                if(!this[prop+'Prop'].includes('full')) size = null
                else size = true
                if(prop !== 'result') {
                    if(index !== 0) {
                        if(this[prop+'Prop'].includes('active')) {
                            hideStyle = ''
                            active = 'tabActiveStyle'
                        } else {
                            hideStyle = `style="display:none;"`
                            active = 'tabStyle'
                        }
                    } else {
                        hideStyle = ''
                        active = 'tabActiveStyle'
                    }
                    tabs1 += `
                        <div contentStyle>
                            <div ${active} onclick="${fuShowHide}">${prop}</div>
                            <div ${hideStyle}>${this.createCodeWithLines(this[prop].trim(),size)}</div>
                        </div>
                    `                
                } else if(prop == 'result') {
                    if(this[prop+'Prop'].includes('active')) {
                        active = 'tabActiveStyle'
                        hideStyle = 'resultStyle'
                    } else {
                        active = 'tabStyle'
                        hideStyle = `resultStyle style="display:none;"`
                    }
                    tabs1 += `
                    <div contentStyle>
                        <div ${active} onclick="${fuShowHide}">Result</div>
                        <div ${hideStyle} id="__result${i}__"></div>
                    </div>
                    `
                }
            });
            tabs1 += `</div>`
            outer = tabs1
        } else outer = this.createCodeWithLines(outer.trim(),null)
        return `<div colorcode id="${dom.id}">${outer}</div>`
    }

    createCodeWithLines(outer,size) {
    // Styles
        if(!this.styleSheet.includes('[prestyle]')) {
            this.styleSheet += `\n
    *,*::before,*::after{box-sizing:border-box}
    [prestyle] {padding:1rem; margin-bottom:0; font-family: monospace; width:100%;}
    [codemainstyle] {
        background-color:#303030; 
        color:white; 
        display:inline-flex; 
        padding-right:.5rem; 
        resize: both; 
        position:relative; 
        max-width: 90vw;
        overflow:auto;
        border: solid white 1px;
        max-width:100%;
    }
    [linestyle] {display:flex; flex-direction:column; font-size:1em; padding:1rem 0 0 1rem; color:gray; font-family: monospace;}
            `
        }
        
        let button = `<button style="position:absolute;" onclick="colorCode.copyText(this.parentNode.children[2])">Copy</button>`
        let lines = outer.split(/\r\n|\r|\n/)
        let linesLength = lines.length
        let linesCol = '<div linestyle>'
        let maxLength = 0
        for (let i = 0; i <= linesLength; i++) {
            if(lines[i] !== undefined) {
                if(lines[i].length>maxLength) maxLength = lines[i].length
            }
            if(i == 0 ) linesCol += `<div>_</div>`
            else linesCol += `<div>${i}</div>`
        }
        let dimentions = ''
        if(size == null) {
            let width = ''
            if(maxLength >300) width = `width:500px;`
            let height = ''
            if(linesLength >16) height = `height: 300px;`
            if(height !== '' || width !== '') dimentions += `style="${width} ${height}"`
        } else dimentions = 'width:inherit; height:inherit;'
        linesCol += '</div>'
        outer = `
        <div codemainstyle ${dimentions}>
            ${button}
            ${linesCol}
            <div><pre prestyle>${outer}</pre></div>
        </div>
        `            
        return outer
    }

    router(inner) {
        if(this.jsProp !== undefined) this.js = ''
        if(this.cssProp !== undefined) this.css = ''

        //this.scriptsToEval = []
        inner = this.getScripts(inner) // get inner without scripts and created this.scripts
        inner = this.getStyles(inner)
        inner = this.colorHtml(inner)

        let scripts = this.scripts
        for(let scriptId in scripts) {
            inner = inner.replace(scriptId,scripts[scriptId])
        }

        let styles = this.styles
        for(let styleId in styles) {
            inner = inner.replace(styleId,styles[styleId])
        }
        return inner
    }

    getScripts(inner) {
// Get Scripts
        let array
        let array2
        let script
        let scriptId
        let scripts = {}
        let self = this
        let originalScript
        if(inner.includes('<script')) {
            array = inner.split('<script')
            array.forEach((e,i) => {
                script = ''
                if(e.includes('</script>')) {
                    originalScript = `<script${e}`
                    e = e.replace('</script>','')
                    array2 = e.split('>')
                    array2.shift()
                    if(array2.length > 1) {
                        array2.forEach(element => {
                            script += '>'+element
                        });
                        script = script.trim()
                    } else script = array2[0].trim()
                    if(script.length>0) {
                        script = script.trim()
                        if(script.charAt(0) =='>') script = script.substring(1);
                        if(self.jsProp !== undefined) {
                            inner = inner.replace(originalScript,'')
                            this.js += script.trim() + '\n'
                        } 
                        else if(scripts[script] == undefined) {
                            scriptId = Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 9)+'script';
                            scripts[scriptId] = script
                            inner = inner.replace(script,scriptId)
                        }
                    }
                    //this.scriptsToEval.push(script)
                }
            });
        }
// Get scripts from actions
        let reg
        let actions = this.actions
        actions.forEach(action => {
            reg = new RegExp(`${action}\s*?=\s*?"`,'g')    
            if(inner.match(reg) !== null) {
                array = inner.match(reg)
                array.forEach(element => {
                    array2 = inner.split(element)
                    array2.shift()
                    array2.forEach(script => {
                        script = script.split('"')[0]
                        if(script.length>0) {
                            scriptId = Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 9);
                            if(scripts[script] == undefined) {
                                scripts[scriptId] = script
                                inner = inner.replace(script,scriptId)
                            }
                        }
                    });
                });
            }
        });
        for(let i in scripts) {
            scripts[i] = this.colorJs(scripts[i])
        }
        if(this.js !== undefined) this.js = this.colorJs(this.js)
        this.scripts = scripts
        return inner
    }

    colorHtml(inner) {
        let outer = ''
// Tags
        let tag
        let firstWord
        let inners = inner.split('<')
        inners.forEach((element,i) => {
            if(i>0) {
                firstWord = element.split(' ')[0]
                if(firstWord.includes('>')) firstWord = firstWord.split('>')[0]
                if(firstWord.startsWith('/')) tag = `&lt;/<span tag-colorcode>`+firstWord.replace('/','')+'</span>'
                else tag = `&lt;<span tag-colorcode>`+firstWord+'</span>'
                element = element.replace(firstWord,tag)
                outer += element
            }
        });
// Props
        let props; let propName; let propValue; let array;
        let actions = this.actions
        props = inner.match(/\w*=\".*?\"/g)
        if(props !== null) {
            props.forEach(prop => {
                array = prop.split('=')
                propName = array[0]
                propValue = array[1]
                propName = prop.replace(propName,`<span prop-colorcode>${propName}</span>`)
                if(propValue == '""') propName = propName.replace('=""','')
                else if(actions.includes(array[0])) propName = propName
                else propName = propName.replace(propValue,`"<span text-colorcode>${propValue.replace('"','').replace('"','')}</span>"`)
                outer = outer.replace(prop,propName)
            });
        }
        outer = this.splitBy('&lt;<span tag-colorcode>!--','-->',outer,'comment')
        //outer = this.htmlComments(outer)
        return outer
    }

    removeQuotesInsideQuotes(script) {
        let quotes = {'"':'&QUOT;',"'":'&apos;','`':'&#96;'}
        let quotesInsideQuotes
        let reg
        let newElement
        for(let quote in quotes) {
            for(let quote1 in quotes) {
                if(quote !== quote1) {
                    reg = new RegExp("\\"+quote+'.*\\'+quote1+'.*\\'+quote+'.*?\n','g') 
                    quotesInsideQuotes = script.match(reg)
                    if(quotesInsideQuotes !== null) {
                        quotesInsideQuotes.forEach(element => {
                            newElement = element
                            while(newElement.includes(quote1)) {
                                newElement = newElement.replace(quote1,quotes[quote1])
                            }
                            script = script.replace(element,newElement)
                        });
                    }
                }
            }
        }
        return script
    }

    colorQuotes(script,quote) {
        let isColored
        let array
        let newElement
        let quotes = script.split(quote)
        quotes.forEach((element,index) => {
            isColored = false
            newElement = ''
            if(index%2 && element !== '') {
                for(let color in this.colors) {
                    if(element.includes(this.colors[color])) {
                        isColored = true
                        break
                    }
                }
                if(!isColored) {
                    if(quote == '`' && element.includes('${')) {
                        array = element.split('${')
                        array.forEach((e,i) => {
                            if(i%2) e = `<span tag-colorcode>$\{</span><span style="color:white;">${e.split('}')[0]}</span><span tag-colorcode>}</span>`+e.split('}')[1]
                            newElement += e
                        });
                    } else newElement = element
        
                    script = script.replace(quote+element+quote,quote+`<span text-colorcode>${newElement}</span>`+quote)
                }
            }
        });
        return script
    }

    colorJs(script) {
        while(script.includes('<')) {
            script = script.replace('<','&lt;')
        }
// Quotes
        script = this.removeQuotesInsideQuotes(script)
        script = this.colorQuotes(script,'"')
        script = this.colorQuotes(script,"'")
        script = this.colorQuotes(script,"`")
// true and false,new,digits
        while(script.includes('false')) {
            script = script.replace('false',`<span fal$%^*@#$!se-colorcode>fal$%^*@#$!se</span>`)
        }
        while(script.includes('null')) {
            script = script.replace('null',`<span false-colorcode>nu$%^*@#$!ll</span>`)
        }
        while(script.includes('undefined')) {
            script = script.replace('undefined',`<span false-colorcode>unde$%^*@#$!find</span>`)
        }
        while(script.includes('class')) {
            script = script.replace('class',`<span let-colorcode>cl$%^*@#$!ass</span>`)
        }
        while(script.includes('this')) {
            script = script.replace('this',`<span th$%^*@#$!is-colorcode>th$%^*@#$!is</span>`)
        }
        while(script.includes('function')) {
            script = script.replace('function',`<span let-colorcode>func$%^*@#$!tion</span>`)
        }
        while(script.includes('else')) {
            script = script.replace('else',`<span tag-colorcode>el$%^*@#$!se</span>`)
        }
        while(script.includes('return')) {
            script = script.replace('return',`<span tag-colorcode>ret$%^*@#$!urn</span>`)
        }
        while(script.includes('typeof')) {
            script = script.replace('typeof',`<span tag-colorcode>type$%^*@#$!of</span>`)
        }
        while(script.includes('==')) {
            script = script.replace('==',`<span tag-colorcode>=$%^*@#$!=</span>`)
        }
        while(script.includes('||')) {
            script = script.replace('||',`<span tag-colorcode>|$%^*@#$!|</span>`)
        }
        while(script.includes('&&')) {
            script = script.replace('&&',`<span tag-colorcode>&$%^*@#$!&</span>`)
        }
        while(script.includes('new')) {
            script = script.replace('new',`<span tag-colorcode>ne$%^*@#$!w</span>`)
        }
        while(script.includes('delete')) {
            script = script.replace('delete',`<span tag-colorcode>dele%^&*@#$!te</span>`)
        }
        while(script.includes('true')) {
            script = script.replace('true',`<span false-colorcode>tr$%^*@#$!ue</span>`)
        }
        while(script.includes('$%^*@#$!')) {
            script = script.replace('$%^*@#$!','')
        }
// const let var
        while(script.includes('let ')) {
            script = script.replace('let ',`<span let-colorcode>let</span> `)
        }
        while(script.includes('const ')) {
            script = script.replace('const ',`<span let-colorcode>const</span> `)
        }
        while(script.includes('var ')) {
            script = script.replace('var ',`<span let-colorcode>var</span> `)
        }
// digits
        let digits = script.match(/\=\s?[0-9].*/g)
        let colorDigit
        if(digits !== null) {
            digits.forEach(digit => {
                colorDigit = digit.replace('=','').trim()
                if(colorDigit !== '') script = script.replace(digit,`= <span false-colorcode>${colorDigit}</span>`)
            });
        }
// functions
        let self = this
        let specials = ['if','for','while']
        let fnName
        let color
        let fns = script.match(/\w*\(/g)
        if(fns !== null) {
            fns.forEach(fn => {
                fnName = fn.replace('(','')
                if(specials.includes(fnName)) color = 'tag-colorcode'
                else color = 'fn-colorcode'
                script = script.replace(fn,`<span ${color}>${fnName}</span>(`)
            });
        }
// Comments
        let comments = script.split('//')
        let newComment
        comments.forEach((comment,index) => {
            if(comment.includes('\n') && comment.split('\n')[0] !== '') {
                comment = comment.split('\n')[0]
                newComment = comment.split('\n')[0]
                newComment = this.removeColors(newComment) 
                newComment = `<span comment-colorcode>//${newComment}</span>`
                script = script.replace('//'+comment,newComment)
            }
        });
        script = this.splitBy('/*','*/',script,'comment')
        return script
    }

    splitBy(first,second,html,color) {
        if(html.includes(first)) {
            let elements = html.split(first)
            let newFirst = this.removeColors(first)
            elements.forEach((element,index) => {
                if(index%2) {
                    element = element.split(second)[0]
                    let newElement = this.removeColors(element)
                    newElement = `<span ${color}-colorcode>${newFirst}${newElement}${second}</span>`
                    html = html.replace(`${first}${element}${second}`,newElement)
                }
            });
        }
        return html
    }

    getStyles(inner) {
// Get Style
        let array
        let array2
        let style
        let styleId
        let styles = {}
        let self = this
        let originalStyle
        if(inner.includes('<style')) {
            array = inner.split('<style>')
            array.forEach((e,i) => {
                style = ''
                if(e.includes('</style>')) {
                    array2 = e.split('</style>')
                    style = e.split('</style>')[0]
                    originalStyle = `<style>${style}</style>`
                    style = style.trim()
                    if(style.length>0) {
                        style = style.trim()
                        style = style.substring(1);
                        if(self.cssProp !== undefined) {
                            inner = inner.replace(originalStyle,'')
                            this.css += style.trim() + '\n'
                        } else if(styles[style] == undefined) {
                            styleId = Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 9)+'style';
                            styles[styleId] = style
                            inner = inner.replace(style,styleId)
                        }
                    }
                }
            });
        }
// Get styles from elements
        let reg
        reg = new RegExp(`style\s*?=\s*?"`,'g')    
        if(inner.match(reg) !== null) {
            array = inner.match(reg)
            array.forEach(element => {
                array2 = inner.split(element)
                array2.shift()
                array2.forEach(style => {
                    style = style.split('"')[0]
                    if(style.length>0) {
                        if(styles[style] == undefined) {
                            styleId = Math.random().toString(36).replace(/[^a-z]+/g, '').substr(0, 9);
                            styles[styleId] = style
                            inner = inner.replace(style,styleId)
                        }
                    }
                });
            });
        }
        if(Object.keys(styles).length > 0) {
            for(let i in styles) {
                styles[i] = this.colorStyles(styles[i])
            }
        }
        if(this.css !== undefined) this.css = this.colorStyles(this.css)
        this.styles = styles
        return inner
    }

    colorStyles(styles) {
        let selectorColor
// get selectors
        let selectors = styles.match(/.*\{/g)
        if(selectors !== null) {
            selectors.forEach(selector => {
                if(selector.includes('@')) { // @media, @keyframes
                    selector = selector.trim().split(' ')[0]
                    styles = styles.replace(selector,`<span tag-colorcode>${selector}</span>`)
                } else {
                    if(selector.includes('.') || selector.includes('#') || selector.includes('[')) {
                        selectorColor = 'prop-colorcode'
                    } else selectorColor = 'tag-colorcode'
                    selector = selector.replace('{','')
                    styles = styles.replace(selector,`<span ${selectorColor}>${selector}</span>`)
                }
            });
        }
// get properties
        let props = styles.match(/.*\:.*/g)
        let propName
        let newProp
        if(props !== null) {
            props.forEach(prop => {
                if(!prop.includes('prop-colorcode') && !prop.includes('/*')) { // for example not [code]:hover
                    if(prop.includes('(') && prop.includes(')')) { //for example @media (max-width:400px) 
                        let insideBrackets = prop.match(/\(.*\)/g)[0]
                        if(insideBrackets !== null) {
                            if(insideBrackets.includes(':')) {
                                propName = insideBrackets.split(':')[0].replace('(','')
                            } else {
                                propName = prop.split(':')[0]
                            }
                            newProp = prop.replace(propName,`<span let-colorcode>${propName}</span>`)
                            styles = styles.replace(prop,newProp)
                        }
                    } else { // if it prop:value
                        let propArray = prop.trim().split(';') // color:red; max-height: 500px;
                        propArray.forEach(propElement => {
                            if(propElement !== '') {
                                styles = styles.replace(propElement,`<span let-colorcode>${propElement.split(':')[0]}</span>:${propElement.split(':')[1]}`)
                            }
                        });
                    }
                }
            });
        }
// Units
        let units = ['px','rem','em','vw','vh','deg','%'] // tag color
        units.forEach(unit => {
            let reg = new RegExp(`[0-9]*${unit}`,'g')
            let match = styles.match(reg)
            if(match !== null) {
                match.forEach(element => {
                    if(element.match(/[0-9]/)) {
                        let newElement = `<span false-colorcode>${element.replace(unit,'')}</span><span tag-colorcode>${unit}</span>`
                        styles = styles.replace(element,newElement)
                    }
                });
            }
        });
// comments
        styles = this.splitBy('/*','*/',styles,'comment')
        return styles
    }

    static copyText(dom) {
        let selection = window.getSelection();
        let range = document.createRange();
        range.selectNodeContents(dom);
        selection.removeAllRanges();
        selection.addRange(range);
        document.execCommand('copy');
    }
   
    removeColors(comment) {
        let spans
        let includesColors = false
        for(let color in this.colors) {
            if(comment.includes(color+'-colorcode')) includesColors = true
        }

        if(comment.includes('span') && includesColors) {
            spans = comment.match(/<span\s\w.*?\>/g)
            spans.forEach(span => {
                comment = comment.replace(span,'')
            });
            while(comment.includes('</span>')) {
                comment = comment.replace('</span>','')
            }
        }
        return comment
    }
}